@@ -1,11 +1,13 @@
 // Take a look at the license at the top of the repository in the LICENSE file.
 
 use crate::sys::component::{self, Component};
+use crate::sys::cpu::*;
 use crate::sys::disk;
 use crate::sys::process::*;
-use crate::sys::processor::*;
-use crate::sys::utils::get_all_data;
-use crate::{Disk, LoadAvg, Networks, Pid, ProcessRefreshKind, RefreshKind, SystemExt, User};
+use crate::sys::utils::{get_all_data, to_u64};
+use crate::{
+    CpuRefreshKind, Disk, LoadAvg, Networks, Pid, ProcessRefreshKind, RefreshKind, SystemExt, User,
+};
 
 use libc::{self, c_char, c_int, sysconf, _SC_CLK_TCK, _SC_HOST_NAME_MAX, _SC_PAGESIZE};
 use std::collections::HashMap;
@@ -61,12 +63,6 @@ pub(crate) fn get_max_nb_fds() -> isize {
     }
 }
 
-macro_rules! to_str {
-    ($e:expr) => {
-        unsafe { std::str::from_utf8_unchecked($e) }
-    };
-}
-
 fn boot_time() -> u64 {
     if let Ok(f) = File::open("/proc/stat") {
         let buf = BufReader::new(f);
@@ -109,7 +105,7 @@ impl SystemInfo {
     fn new() -> Self {
         unsafe {
             Self {
-                page_size_kb: (sysconf(_SC_PAGESIZE) / 1024) as _,
+                page_size_kb: sysconf(_SC_PAGESIZE) as _,
                 clock_cycle: sysconf(_SC_CLK_TCK) as _,
                 boot_time: boot_time(),
             }
@@ -161,21 +157,16 @@ pub struct System {
     mem_available: u64,
     mem_buffers: u64,
     mem_page_cache: u64,
+    mem_shmem: u64,
     mem_slab_reclaimable: u64,
     swap_total: u64,
     swap_free: u64,
-    global_processor: Processor,
-    processors: Vec<Processor>,
     components: Vec<Component>,
     disks: Vec<Disk>,
     networks: Networks,
     users: Vec<User>,
-    /// Field set to `false` in `update_processors` and to `true` in `refresh_processes_specifics`.
-    ///
-    /// The reason behind this is to avoid calling the `update_processors` more than necessary.
-    /// For example when running `refresh_all` or `refresh_specifics`.
-    need_processors_update: bool,
     info: SystemInfo,
+    cpus: CpusWrapper,
 }
 
 impl System {
@@ -185,23 +176,22 @@ impl System {
     /// To prevent that, we compute ahead of time this maximum value and ensure that processes'
     /// CPU usage don't go over it.
     fn get_max_process_cpu_usage(&self) -> f32 {
-        self.processors.len() as f32 * 100.
+        self.cpus.len() as f32 * 100.
     }
 
     fn clear_procs(&mut self, refresh_kind: ProcessRefreshKind) {
         let (total_time, compute_cpu, max_value) = if refresh_kind.cpu() {
-            if self.need_processors_update {
-                self.refresh_processors(true);
-            }
+            self.cpus
+                .refresh_if_needed(true, CpuRefreshKind::new().with_cpu_usage());
 
-            if self.processors.is_empty() {
-                sysinfo_debug!("cannot compute processes CPU usage: no processor found...");
+            if self.cpus.is_empty() {
+                sysinfo_debug!("cannot compute processes CPU usage: no CPU found...");
                 (0., false, 0.)
             } else {
-                let (new, old) = get_raw_times(&self.global_processor);
+                let (new, old) = self.cpus.get_global_raw_times();
                 let total_time = if old > new { 1 } else { new - old };
                 (
-                    total_time as f32 / self.processors.len() as f32,
+                    total_time as f32 / self.cpus.len() as f32,
                     true,
                     self.get_max_process_cpu_usage(),
                 )
@@ -222,124 +212,8 @@ impl System {
         });
     }
 
-    fn refresh_processors(&mut self, only_update_global_processor: bool) {
-        if let Ok(f) = File::open("/proc/stat") {
-            self.need_processors_update = false;
-
-            let buf = BufReader::new(f);
-            let mut i: usize = 0;
-            let first = self.processors.is_empty();
-            let mut it = buf.split(b'\n');
-            let (vendor_id, brand) = if first {
-                get_vendor_id_and_brand()
-            } else {
-                (String::new(), String::new())
-            };
-
-            if let Some(Ok(line)) = it.next() {
-                if &line[..4] != b"cpu " {
-                    return;
-                }
-                let mut parts = line.split(|x| *x == b' ').filter(|s| !s.is_empty());
-                if first {
-                    self.global_processor.name = to_str!(parts.next().unwrap_or(&[])).to_owned();
-                } else {
-                    parts.next();
-                }
-                self.global_processor.set(
-                    parts.next().map(to_u64).unwrap_or(0),
-                    parts.next().map(to_u64).unwrap_or(0),
-                    parts.next().map(to_u64).unwrap_or(0),
-                    parts.next().map(to_u64).unwrap_or(0),
-                    parts.next().map(to_u64).unwrap_or(0),
-                    parts.next().map(to_u64).unwrap_or(0),
-                    parts.next().map(to_u64).unwrap_or(0),
-                    parts.next().map(to_u64).unwrap_or(0),
-                    parts.next().map(to_u64).unwrap_or(0),
-                    parts.next().map(to_u64).unwrap_or(0),
-                );
-                if !first && only_update_global_processor {
-                    return;
-                }
-            }
-            while let Some(Ok(line)) = it.next() {
-                if &line[..3] != b"cpu" {
-                    break;
-                }
-
-                let mut parts = line.split(|x| *x == b' ').filter(|s| !s.is_empty());
-                if first {
-                    self.processors.push(Processor::new_with_values(
-                        to_str!(parts.next().unwrap_or(&[])),
-                        parts.next().map(to_u64).unwrap_or(0),
-                        parts.next().map(to_u64).unwrap_or(0),
-                        parts.next().map(to_u64).unwrap_or(0),
-                        parts.next().map(to_u64).unwrap_or(0),
-                        parts.next().map(to_u64).unwrap_or(0),
-                        parts.next().map(to_u64).unwrap_or(0),
-                        parts.next().map(to_u64).unwrap_or(0),
-                        parts.next().map(to_u64).unwrap_or(0),
-                        parts.next().map(to_u64).unwrap_or(0),
-                        parts.next().map(to_u64).unwrap_or(0),
-                        0,
-                        vendor_id.clone(),
-                        brand.clone(),
-                    ));
-                } else {
-                    parts.next(); // we don't want the name again
-                    self.processors[i].set(
-                        parts.next().map(to_u64).unwrap_or(0),
-                        parts.next().map(to_u64).unwrap_or(0),
-                        parts.next().map(to_u64).unwrap_or(0),
-                        parts.next().map(to_u64).unwrap_or(0),
-                        parts.next().map(to_u64).unwrap_or(0),
-                        parts.next().map(to_u64).unwrap_or(0),
-                        parts.next().map(to_u64).unwrap_or(0),
-                        parts.next().map(to_u64).unwrap_or(0),
-                        parts.next().map(to_u64).unwrap_or(0),
-                        parts.next().map(to_u64).unwrap_or(0),
-                    );
-                }
-
-                i += 1;
-            }
-
-            #[cfg(feature = "multithread")]
-            use rayon::iter::{
-                IndexedParallelIterator, IntoParallelRefMutIterator, ParallelIterator,
-            };
-
-            #[cfg(feature = "multithread")]
-            // This function is voluntarily made generic in case we want to generalize it.
-            fn iter_mut<'a, T>(
-                val: &'a mut T,
-            ) -> <&'a mut T as rayon::iter::IntoParallelIterator>::Iter
-            where
-                &'a mut T: rayon::iter::IntoParallelIterator,
-            {
-                val.par_iter_mut()
-            }
-
-            #[cfg(not(feature = "multithread"))]
-            fn iter_mut<'a>(val: &'a mut Vec<Processor>) -> std::slice::IterMut<'a, Processor> {
-                val.iter_mut()
-            }
-
-            // `get_cpu_frequency` is very slow, so better run it in parallel.
-            self.global_processor.frequency = iter_mut(&mut self.processors)
-                .enumerate()
-                .map(|(pos, proc_)| {
-                    proc_.frequency = get_cpu_frequency(pos);
-                    proc_.frequency
-                })
-                .max()
-                .unwrap_or(0);
-
-            if first {
-                self.global_processor.vendor_id = vendor_id;
-                self.global_processor.brand = brand;
-            }
-        }
+    fn refresh_cpus(&mut self, only_update_global_cpu: bool, refresh_kind: CpuRefreshKind) {
+        self.cpus.refresh(only_update_global_cpu, refresh_kind);
     }
 }
 
@@ -348,8 +222,7 @@ impl SystemExt for System {
     const SUPPORTED_SIGNALS: &'static [Signal] = supported_signals();
 
     fn new_with_specifics(refreshes: RefreshKind) -> System {
-        let info = SystemInfo::new();
-        let process_list = Process::new(Pid(0), None, 0, &info);
+        let process_list = Process::new(Pid(0));
         let mut s = System {
             process_list,
             mem_total: 0,
@@ -357,32 +230,16 @@ impl SystemExt for System {
             mem_available: 0,
             mem_buffers: 0,
             mem_page_cache: 0,
+            mem_shmem: 0,
             mem_slab_reclaimable: 0,
             swap_total: 0,
             swap_free: 0,
-            global_processor: Processor::new_with_values(
-                "",
-                0,
-                0,
-                0,
-                0,
-                0,
-                0,
-                0,
-                0,
-                0,
-                0,
-                0,
-                String::new(),
-                String::new(),
-            ),
-            processors: Vec::with_capacity(4),
+            cpus: CpusWrapper::new(),
             components: Vec::new(),
             disks: Vec::with_capacity(2),
             networks: Networks::new(),
             users: Vec::new(),
-            need_processors_update: true,
-            info,
+            info: SystemInfo::new(),
         };
         s.refresh_specifics(refreshes);
         s
@@ -394,14 +251,20 @@ impl SystemExt for System {
 
     fn refresh_memory(&mut self) {
         if let Ok(data) = get_all_data("/proc/meminfo", 16_385) {
+            let mut mem_available_found = false;
+
             for line in data.split('\n') {
                 let mut iter = line.split(':');
                 let field = match iter.next() {
                     Some("MemTotal") => &mut self.mem_total,
                     Some("MemFree") => &mut self.mem_free,
-                    Some("MemAvailable") => &mut self.mem_available,
+                    Some("MemAvailable") => {
+                        mem_available_found = true;
+                        &mut self.mem_available
+                    }
                     Some("Buffers") => &mut self.mem_buffers,
                     Some("Cached") => &mut self.mem_page_cache,
+                    Some("Shmem") => &mut self.mem_shmem,
                     Some("SReclaimable") => &mut self.mem_slab_reclaimable,
                     Some("SwapTotal") => &mut self.swap_total,
                     Some("SwapFree") => &mut self.swap_free,
@@ -410,15 +273,26 @@ impl SystemExt for System {
                 if let Some(val_str) = iter.next().and_then(|s| s.trim_start().split(' ').next()) {
                     if let Ok(value) = u64::from_str(val_str) {
                         // /proc/meminfo reports KiB, though it says "kB". Convert it.
-                        *field = value * 128 / 125;
+                        *field = value.saturating_mul(1_024);
                     }
                 }
             }
+
+            // Linux < 3.14 may not have MemAvailable in /proc/meminfo
+            // So it should fallback to the old way of estimating available memory
+            // https://github.com/KittyKatt/screenFetch/issues/386#issuecomment-249312716
+            if !mem_available_found {
+                self.mem_available = self.mem_free
+                    + self.mem_buffers
+                    + self.mem_page_cache
+                    + self.mem_slab_reclaimable
+                    - self.mem_shmem;
+            }
         }
     }
 
-    fn refresh_cpu(&mut self) {
-        self.refresh_processors(false);
+    fn refresh_cpu_specifics(&mut self, refresh_kind: CpuRefreshKind) {
+        self.refresh_cpus(false, refresh_kind);
     }
 
     fn refresh_processes_specifics(&mut self, refresh_kind: ProcessRefreshKind) {
@@ -432,7 +306,7 @@ impl SystemExt for System {
             refresh_kind,
         );
         self.clear_procs(refresh_kind);
-        self.need_processors_update = true;
+        self.cpus.set_need_cpus_update();
     }
 
     fn refresh_process_specifics(&mut self, pid: Pid, refresh_kind: ProcessRefreshKind) -> bool {
@@ -454,18 +328,18 @@ impl SystemExt for System {
         };
         if found {
             if refresh_kind.cpu() {
-                self.refresh_processors(true);
+                self.refresh_cpus(true, CpuRefreshKind::new().with_cpu_usage());
 
-                if self.processors.is_empty() {
-                    sysinfo_debug!("Cannot compute process CPU usage: no processors found...");
+                if self.cpus.is_empty() {
+                    sysinfo_debug!("Cannot compute process CPU usage: no cpus found...");
                     return found;
                 }
-                let (new, old) = get_raw_times(&self.global_processor);
+                let (new, old) = self.cpus.get_global_raw_times();
                 let total_time = (if old >= new { 1 } else { new - old }) as f32;
 
                 let max_cpu_usage = self.get_max_process_cpu_usage();
                 if let Some(p) = self.process_list.tasks.get_mut(&pid) {
-                    compute_cpu_usage(p, total_time / self.processors.len() as f32, max_cpu_usage);
+                    compute_cpu_usage(p, total_time / self.cpus.len() as f32, max_cpu_usage);
                     p.updated = false;
                 }
             } else if let Some(p) = self.process_list.tasks.get_mut(&pid) {
@@ -503,12 +377,12 @@ impl SystemExt for System {
         &mut self.networks
     }
 
-    fn global_processor_info(&self) -> &Processor {
-        &self.global_processor
+    fn global_cpu_info(&self) -> &Cpu {
+        &self.cpus.global_cpu
     }
 
-    fn processors(&self) -> &[Processor] {
-        &self.processors
+    fn cpus(&self) -> &[Cpu] {
+        &self.cpus.cpus
     }
 
     fn physical_core_count(&self) -> Option<usize> {
@@ -528,11 +402,7 @@ impl SystemExt for System {
     }
 
     fn used_memory(&self) -> u64 {
-        self.mem_total
-            - self.mem_free
-            - self.mem_buffers
-            - self.mem_page_cache
-            - self.mem_slab_reclaimable
+        self.mem_total - self.mem_available
     }
 
     fn total_swap(&self) -> u64 {
@@ -564,6 +434,13 @@ impl SystemExt for System {
         &mut self.disks
     }
 
+    fn sort_disks_by<F>(&mut self, compare: F)
+    where
+        F: FnMut(&Disk, &Disk) -> std::cmp::Ordering,
+    {
+        self.disks.sort_unstable_by(compare);
+    }
+
     fn uptime(&self) -> u64 {
         let content = get_all_data("/proc/uptime", 50).unwrap_or_default();
         content
@@ -682,6 +559,25 @@ impl SystemExt for System {
     fn os_version(&self) -> Option<String> {
         get_system_info_android(InfoType::OsVersion)
     }
+
+    #[cfg(not(target_os = "android"))]
+    fn distribution_id(&self) -> String {
+        get_system_info_linux(
+            InfoType::DistributionID,
+            Path::new("/etc/os-release"),
+            Path::new(""),
+        )
+        .unwrap_or_else(|| std::env::consts::OS.to_owned())
+    }
+
+    #[cfg(target_os = "android")]
+    fn distribution_id(&self) -> String {
+        // Currently get_system_info_android doesn't support InfoType::DistributionID and always
+        // returns None. This call is done anyway for consistency with non-Android implementation
+        // and to suppress dead-code warning for DistributionID on Android.
+        get_system_info_android(InfoType::DistributionID)
+            .unwrap_or_else(|| std::env::consts::OS.to_owned())
+    }
 }
 
 impl Default for System {
@@ -690,23 +586,16 @@ impl Default for System {
     }
 }
 
-fn to_u64(v: &[u8]) -> u64 {
-    let mut x = 0;
-
-    for c in v {
-        x *= 10;
-        x += u64::from(c - b'0');
-    }
-    x
-}
-
-#[derive(PartialEq)]
+#[derive(PartialEq, Eq)]
 enum InfoType {
     /// The end-user friendly name of:
     /// - Android: The device model
     /// - Linux: The distributions name
     Name,
     OsVersion,
+    /// Machine-parseable ID of a distribution, see
+    /// https://www.freedesktop.org/software/systemd/man/os-release.html#ID=
+    DistributionID,
 }
 
 #[cfg(not(target_os = "android"))]
@@ -717,6 +606,7 @@ fn get_system_info_linux(info: InfoType, path: &Path, fallback_path: &Path) -> O
         let info_str = match info {
             InfoType::Name => "NAME=",
             InfoType::OsVersion => "VERSION_ID=",
+            InfoType::DistributionID => "ID=",
         };
 
         for line in reader.lines().flatten() {
@@ -735,6 +625,10 @@ fn get_system_info_linux(info: InfoType, path: &Path, fallback_path: &Path) -> O
     let info_str = match info {
         InfoType::OsVersion => "DISTRIB_RELEASE=",
         InfoType::Name => "DISTRIB_ID=",
+        InfoType::DistributionID => {
+            // lsb-release is inconsistent with os-release and unsupported.
+            return None;
+        }
     };
     for line in reader.lines().flatten() {
         if let Some(stripped) = line.strip_prefix(info_str) {
@@ -750,6 +644,10 @@ fn get_system_info_android(info: InfoType) -> Option<String> {
     let name: &'static [u8] = match info {
         InfoType::Name => b"ro.product.model\0",
         InfoType::OsVersion => b"ro.build.version.release\0",
+        InfoType::DistributionID => {
+            // Not supported.
+            return None;
+        }
     };
 
     let mut value_buffer = vec![0u8; libc::PROP_VALUE_MAX as usize];
@@ -783,6 +681,7 @@ mod test {
     fn lsb_release_fallback_android() {
         assert!(get_system_info_android(InfoType::OsVersion).is_some());
         assert!(get_system_info_android(InfoType::Name).is_some());
+        assert!(get_system_info_android(InfoType::DistributionID).is_none());
     }
 
     #[test]
@@ -829,6 +728,10 @@ DISTRIB_DESCRIPTION="Ubuntu 20.10"
             get_system_info_linux(InfoType::Name, &tmp1, Path::new("")),
             Some("Ubuntu".to_owned())
         );
+        assert_eq!(
+            get_system_info_linux(InfoType::DistributionID, &tmp1, Path::new("")),
+            Some("ubuntu".to_owned())
+        );
 
         // Check for the "fallback" path: "/etc/lsb-release"
         assert_eq!(
@@ -839,5 +742,9 @@ DISTRIB_DESCRIPTION="Ubuntu 20.10"
             get_system_info_linux(InfoType::Name, Path::new(""), &tmp2),
             Some("Ubuntu".to_owned())
         );
+        assert_eq!(
+            get_system_info_linux(InfoType::DistributionID, Path::new(""), &tmp2),
+            None
+        );
     }
 }
